/* Virtuino SE Json example: How to change the WiFi SSID, WiFi password, server IP and server Gateway of the ESP board
 * using the Virtuino SE app (without compile the code)
 * This is professional code for shared projects. 
 * The users can change the board settings without compile the code using the Virtuino SE app
 * Supported boards: ESP32, ESP8266
 * Created by Ilias Lamprou
 * Updated May 12 2019
 * More examples: https://virtuino.com/index.php/virtuino-se/20-virtuino-se-code-examples
 */
// ------ VirtuinoSE V Memory Table
//  V0 memory -> digital output pin LED_BUILTIN
//  V1 memory -> reserved
//  V2 memory -> variable 
//  V3 memory -> Sensor1 
//  V3 memory -> Sensor2 
//  V4 memory -> A0  analog input
//  V5 memory -> reserved
//  V6 ... V15 memory -> reserved



#define outPin_1 LED_BUILTIN
#ifdef ESP8266
 #include <ESP8266WiFi.h>  // Pins for board ESP8266 Wemos-NodeMCU
 #define accessPointButtonPin D6   // Connect a button to this pin
 #else
 #include <WiFi.h>    // Pins for board ESP32
 #define accessPointButtonPin 4    // Connect a button to this pin
 #endif
    

//------------------- USER SETTINGS ---------------------------------
String ssid = "WIFI NETWORK";           // The name of your WiFi network (SSID)
String password = "PASSWORD";           // WIFI network PASSWORD
String serverIP="192.168.1.150";        // eeprom variable 3
String serverGateway="192.168.1.1";     // eeprom variable 4    


WiFiServer wifiServer(8000);                  // Virtuino default Server port: 8000

const char* ssid_AP = "Virtuino network";     // Access point network SSID
const char* password_AP = "1234567890";       // Access point network PASSWORD



#define VIRTUINO_KEY  "1234"      // Virtuino KEY, only the requests that include this key will be acceptable. // disable the line if you don't want to use a key                          

#define V_memory_count 16         // the size of V memory. You can change it to a number <256)
//-------------------------------------------------------------------

float V_memory[V_memory_count];    // This array is synchronized with Virtuino SE V memory

boolean debug = true;              // set this variable to false on the finale code to decrease the request duration

//--- VARIABLES for this demo example
float   my_Variable = 0;
byte inputPinState=0;



//--- EEPROM VARIABLES for this demo example
#define eepromVariablesCount 5         // the number of variables you want to store to eeprom.  
                                       // eppromVariable 0 is a system variable
                                       // eppromVariable 1 is the wifi ssid
                                       // eppromVariable 2 is the wifi password
                                       // eppromVariable 3 is the server ip
                                       // eppromVariable 4 is the gateway
                                       
#define eepromVariableSize 20          // the max size of each variable. The code divides the EEPRPM memory to parts of 20 chars
                                       // each part is an eepron text variable

boolean isAccessPointMode=false;   // on Void setup, set this variable to true to start the board as access point
unsigned long timeStrored=0;


//============================================================== connectToWiFiNetwork
void connectToWiFiNetwork(){
   if (debug) Serial.println("Connecting to "+String(ssid));
  
   // If you don't want to config IP manually disable the next four lines
  IPAddress ip=getIpFromString(serverIP);
  IPAddress gateway=getIpFromString(serverGateway);         // set gateway to match your network
  IPAddress subnet(255, 255, 255, 0);        // set subnet mask to match your network
  WiFi.config(ip, gateway, subnet);          // If you don't want to config IP manually disable this line
 
  WiFi.mode(WIFI_STA);                       // Configure the module as station only.
  
  char ssid_[eepromVariableSize]; 
  ssid.toCharArray(ssid_,eepromVariableSize);
  char password_[eepromVariableSize]; 
  password.toCharArray(password_,eepromVariableSize);
 
   WiFi.begin(ssid_, password_);
   while (WiFi.status() != WL_CONNECTED) {
     delay(500);
      if (debug) Serial.print(".");
    }
    if (debug) Serial.println("");
    if (debug) Serial.println("WiFi connected");
    if (debug) Serial.println(WiFi.localIP());
}

//================================================================= initAccessPoint
void initAccessPoint(){
   if (debug) Serial.print("Setting soft-AP ... ");                   // Default IP: 192.168.4.1
   WiFi.mode(WIFI_AP);                                     // Config module as Access point only.  Set WiFi.mode(WIFI_AP_STA); to config module as Acces point and station
   boolean result = WiFi.softAP(ssid_AP,password_AP);      // set the Access point SSID and password
   if(result == true)  {
     if (debug)Serial.println("Server Ready");
     if (debug) Serial.println(WiFi.softAPIP());
   }
   else  if (debug)Serial.println("Failed!");
}

//============================================================== setup
//==============================================================
void setup() {
  if (debug) {
    Serial.begin(9600);
    while (!Serial) continue;
  }

  pinMode(accessPointButtonPin,INPUT);    // Push the conected to "accessPointButtonPin" button while the board is starting to start the board as access point
  if (digitalRead(accessPointButtonPin)==HIGH) isAccessPointMode=true;          
  else isAccessPointMode=false;          // set to true every time the board is started as access point
                                    // use a button to select the access point mode after a board restarting  
 
  //writeDefaultSettingsToEEPROM(); // enable this line every time you want to save the default settings to EEPROM
                                    // you have to enable it the first time you will upload this code and then you have to disable it
  readSettingsFromEEPROM();         // disable this line if you want to start the board with the default settings

  if (isAccessPointMode) initAccessPoint(); // start the board as access point or as virtuino server (Json API code)
  else connectToWiFiNetwork();      
  
  wifiServer.begin();
  
  pinMode(outPin_1, OUTPUT);            
 
 }




//============================================================== loop
//==============================================================
void loop() {
  virtuinoRun();        //Necessary function to communicate with Virtuino. Client handler

  //--- How to control outputs from Virtuino SE
    if (vMemoryRead(0)==1) digitalWrite(outPin_1,HIGH);      // On Virtuino SE panel add switch to control the pin outPin_1. On the switch settings select the memory V0
    else digitalWrite(outPin_1,LOW);
  
 
  //--- How to read a variable from Virtuino SE
  if (my_Variable!=vMemoryRead(2)) {        // Read the V2 memory.  On Virtuino panel add a regulator  to virtual memory V2  
      my_Variable==vMemoryRead(2);
  }
  
  //--- How to send to Virtuino SE sensor values
   if (millis()-timeStrored>5000) {       // read sensor value every 5 seconds. Avoid to use delay on your code
      vMemoryWrite(3,random(100));           // copy the sensor value to V5
      vMemoryWrite(4,random(100));          // copy the sensor value to V6
      vMemoryWrite(5,analogRead(A0));       // copy the A0 value V7
      timeStrored=millis();               // On Virtuino panel add value displays or instrument  to virtual memories V3,V4,V5
  }


}


//===========================================================
float vMemoryRead(int memoryIndex){
  if ((memoryIndex>=0) & (memoryIndex<V_memory_count)) return V_memory[memoryIndex];
  else return 0;
}
//===========================================================
void vMemoryWrite(int memoryIndex, double value){
  if ((memoryIndex,0)|| (memoryIndex>=V_memory_count)) return; 
  else V_memory[memoryIndex]=value;
}



//============================================================== onCommandReceived
//==============================================================
/* This function is called every time Virtuino SE app sends a request to server 
 * The 'variableIndex' is the V memory index of Virtuino app
 * The 'valueAsText' is the value that has sent from the app
 * If the 'valueAsText" is ? you have to return the value that you want to be displayed on Virtuino panel
 */
 String onCommandReceived(int variableIndex, String valueAsJson){     
   String valueAsText = getJsonValue(valueAsJson,"value");    // read the value from the received Json string 
   if (debug) Serial.println("New value received: V"+String(variableIndex)+"  value="+valueAsText);
  if (isAccessPointMode) {    // AccessPoint code
   if (valueAsText=="?") {
    if (variableIndex==10) return ssid;
    else if (variableIndex==11) return password;
     else if (variableIndex==12) return serverIP;
       else if (variableIndex==13) return serverGateway;
   }
 
  else {
    if (variableIndex==10) {
      writeToEeprom(1,valueAsText);
      ssid=readFromEeprom(1);   // read the saved value
    }
    else if (variableIndex==11) {
      writeToEeprom(2,valueAsText);
      password=readFromEeprom(2);   // read the saved value
    }
    else if (variableIndex==12) {
      writeToEeprom(3,valueAsText);
      serverIP=readFromEeprom(3);   // read the saved value
    }
    else if (variableIndex==13) {
      writeToEeprom(4,valueAsText);
      serverGateway=readFromEeprom(4);   // read the saved value
    }
    
  }
   } //if (isAccesPointMode

   else 
      {    // default client JSON API code 
          if (valueAsText=="?") return (String(vMemoryRead(variableIndex)));   // return the value of the arduino V memory array
          else {
                  float value = valueAsText.toFloat();        // convert the value to float. The valueAsText have to be numerical
                  vMemoryWrite(variableIndex,value);
                }
   
      }
   
  
  return "";
}










//================== Virtuino Json API Library (included in sketch) ===============
//==============================================================
//==============================================================
/* Only the next functions are nedded to communicate with Virtuino 
 * Avoid to make changes to the code below  
 */


#define WRONG_KEY_MESSAGE  "Wrong Key"  
#define WELLCOME_MESSAGE  "Hello Virtuino"  


//============================================================== virtuinoRun
//==============================================================
  void virtuinoRun(){
  WiFiClient client = wifiServer.available();
  if (client) {
    if (debug) Serial.println("Client connected");
    while(!client.available())delay(1);    // Wait until the client sends some data
    String data="";
    while (client.connected()) {
      while (client.available()>0) {
        char c = client.read();
        data+=c;
        if (debug) Serial.write(c);
      }
     client.flush(); 
     if (debug) Serial.println("\nReceived data: "+data);
     String response= getResponseAsJson(data);
      client.print(response);
     if (debug) Serial.println("Response : "+response);
     client.stop();
    }
     
    if (debug) Serial.println("Client disconnected");
  }
 }



//============================================================== prepareResponse
//==============================================================
String getResponseAsJson(String data){
    //---check the KEY
   #ifdef VIRTUINO_KEY 
      String key = getJsonValue(data, "key");
      if (debug) Serial.println("Key="+key);
      if (!key.equals(VIRTUINO_KEY)) return "{\"status\":\"-1\",\"message\":\""+String(WRONG_KEY_MESSAGE)+"\"}";
   #endif

    //---check connection status
    String statusInfo = getJsonValue(data, "status");
    if (debug) Serial.println("statusInfo="+statusInfo);
    if (statusInfo.equals("0")) return "{\"status\":\"2\",\"message\":\""+String(WELLCOME_MESSAGE)+"\"}";
  
   String response="{\"status\":\"1\"";
   for (byte i=0;i<V_memory_count;i++){
      // easy way to read a Json object without installing extra json libraries
      String jsonTag="V"+String(i)+"_";
      int pos = data.indexOf(jsonTag);
      if (pos>0) {
        int startPos= data.indexOf("{",pos+1);
        if (startPos>pos) {
          int endPos= data.indexOf("}", startPos+2);
          if (endPos>startPos) {
            String variableAsJson = data.substring(startPos,endPos+1);
            String variableValue=onCommandReceived(i,variableAsJson);
           if (variableValue.length()>0) response+=",\""+jsonTag+"\":{\"value\":\""+urlencode(&variableValue)+"\"}";
          }
        }
      }
      
   } //for
   response+="}";
  return response;
}

//==============================================================
// easy way to read a Json text value without installing extra json libraries
  String getJsonValue(String jsonAsText, String jsonTag){
    int pos = jsonAsText.indexOf(jsonTag);
    if (pos>0) {
      int startPos=jsonAsText.indexOf(":\"",pos);
      int lastPos = jsonAsText.indexOf("\"",startPos+2);
      String value=jsonAsText.substring(startPos+2,lastPos);
      value=urldecode(&value);
      if (lastPos>startPos) return value;
    }
    return "";
 }


  //=================== ENCODE - DECODE Library (included in sketch) ==========
  //=================================================================
  
  //================================================================ urldecode
String urldecode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    for (int i =0; i < str->length(); i++){
        c=str->charAt(i);
      if (c == '+'){
        encodedString+=' ';  
      }else if (c == '%') {
        i++;
        code0=str->charAt(i);
        i++;
        code1=str->charAt(i);
        c = (h2int(code0) << 4) | h2int(code1);
        encodedString+=c;
      } else{
         encodedString+=c;  
      }
      yield();
    }
    
   return encodedString;
}
//================================================================ urlencode
String urlencode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    char code2;
    for (int i =0; i < str->length(); i++){
      c=str->charAt(i);
      if (c == ' '){
        encodedString+= '+';
      } else if (isalnum(c)){
        encodedString+=c;
      } else{
        code1=(c & 0xf)+'0';
        if ((c & 0xf) >9){
            code1=(c & 0xf) - 10 + 'A';
        }
        c=(c>>4)&0xf;
        code0=c+'0';
        if (c > 9){
            code0=c - 10 + 'A';
        }
        code2='\0';
        encodedString+='%';
        encodedString+=code0;
        encodedString+=code1;
        //encodedString+=code2;
      }
      yield();
    }
    return encodedString;
}
//================================================================ h2int
unsigned char h2int(char c){
    if (c >= '0' && c <='9'){
        return((unsigned char)c - '0');
    }
    if (c >= 'a' && c <='f'){
        return((unsigned char)c - 'a' + 10);
    }
    if (c >= 'A' && c <='F'){
        return((unsigned char)c - 'A' + 10);
    }
    return(0);
}





//============================= EEPROM code ======================
//================================================================
//================================================================
#include <EEPROM.h>
int eepromSize=eepromVariablesCount*eepromVariableSize;

// EEPROM variables index
// EEPROM index 0 general settings
// EEPROM index 1 WiFi SSID
// EEPROM index 2 WiFi Password
// EEPROM index 3 Server IP
// EEPROM index 4 Gateway IP
 
//========================================== writeDefaultSettingsToEEPROM
void writeDefaultSettingsToEEPROM(){
    writeToEeprom(0,"90");                  // if the first character of the eepron is not 9 you have to write the default settings to eeprom (first run)
    writeToEeprom(1,ssid);    
    writeToEeprom(2,password);
    writeToEeprom(3,serverIP);
    writeToEeprom(4,serverGateway);
 
    
}
//========================================== readSettingsFromEeprom
void readSettingsFromEEPROM(){
  String eepromStartPart = readFromEeprom(0);
  if (eepromStartPart.indexOf("9")!=0) writeDefaultSettingsToEEPROM();
  
  ssid=readFromEeprom(1);
  password=readFromEeprom(2);
  serverIP=readFromEeprom(3);
  serverGateway=readFromEeprom(4);
}


//================================================================ readEepromSettings
  String readFromEeprom(byte index){
   String s="";
   EEPROM.begin(eepromSize);
   int adr=index*eepromVariableSize;
   byte b= EEPROM.read(adr); 
   if (b!=0) {
      char d[eepromVariableSize];
      EEPROM.get (adr, d);
      d[eepromVariableSize-1]=0;
      s=d;
   }
   EEPROM.end();
   return s;
 } 


 //================================================================ writeEepromwifi_SSID
 void writeToEeprom(byte index,String s){
   EEPROM.begin(eepromSize);
   int adr=index*eepromVariableSize;
   for (int i = 0; i < eepromVariableSize; i++ ) {
        if (i<s.length()) EEPROM.write(adr+i, s.charAt(i));
        else EEPROM.write(adr+i,'\0');
        }
   EEPROM.end();     
 }

//===============================================================
IPAddress getIpFromString(String ipString){
  int pos1 = 0;
  int pos2 = ipString.indexOf('.',pos1);
  String part1 = ipString.substring(pos1,pos2);
  pos1 = pos2+1;
  pos2 = ipString.indexOf('.',pos1);
  String part2 = ipString.substring(pos1,pos2);
  pos1 = pos2+1;
  pos2 = ipString.indexOf('.',pos1);
  String part3 = ipString.substring(pos1,pos2);
   pos1 = pos2+1;
  String part4 = ipString.substring(pos1);
  IPAddress ip(part1.toInt(),part2.toInt(),part3.toInt(),part4.toInt());    
  return ip;
  
}



